# Image to High Efficiency Image (AVIF/HIF) Converter 

from PIL import Image
import piexif
import pillow_heif

import heiya.extensions as extensions
import heiya.tools as tools

import os
from os import listdir
from os.path import dirname, basename


def convert_image_to_hei(source_image, target_format=".AVIF"):
    """
    Convert a TIF file into a AVIF file.
    This conversion preserves all the image metadata.
    
    Args:
        source_tif (str): A full file path of a .TIF image.
        target_format (str): The target format you want to convert to, by default use ".HIF"
        
    Returns:
        (str) Full file path of the generated file.
    """
    # Separate a full file path into directory, file name, and extension
    directory = dirname(source_image)
    file_name  = basename(source_image).split(".")[0] 
    extension = basename(source_image).split(".")[1]

    # Register the pillow HEI opener
    if target_format in extensions.EXT_HIF:
        pillow_heif.register_heif_opener()
    elif target_format in extensions.EXT_AVIF:
        pillow_heif.register_avif_opener()
    else:
        raise ValueError("Not a valid target format. Please use .HIF or .AVIF.")

    # Open the image using PIL
    img = Image.open(source_image)

    # Load meta data from tif
    meta_dict = piexif.load(source_image)

    # Convert meta data into bytes so we can pass it into img.save()
    meta = piexif.dump(meta_dict)

    # Construct output file name
    output_file = directory + "/" + file_name + target_format

    # Export file using the meta data information
    img.save(output_file, exif=meta)
    
    return output_file
 
    
def convert_image_in_dir_to_hei(source_dir, source_tif=False, source_jpg=True, target_hif=False, target_avif=False):
    """
    Convert all the files with an extension of ".tif" into target format.
    Args:
        source_dir (str): A directory that contain tif files.
        target_format (str or tuple): The target format you want to convert to, by default use ".HIF"
    """
    try:
        # Filter out hidden cache files starts with "._" created by Capture One
        source_list = []
        
        if source_tif:
            source_list.extend(list(extensions.EXT_TIF))

        if source_jpg:
            source_list.extend(list(extensions.EXT_JPG))
            
        source_format = tuple(source_list)

        source_file_list = [file for file in listdir(source_dir) if file.endswith(source_format) and not file.startswith("._")]  

        # Keep track of image count for displaying progress
        image_count = len(source_file_list)
        image_counter = 0

        # Manipulate each tif image
        for source_file in source_file_list:
            source_file_path = os.path.join(source_dir, source_file)
            
            image_counter += 1
            progress = str(image_counter) + "/" + str(image_count) + "(" + str(int((image_counter / image_count)*100)) + "%)"
            
            if target_hif:
                output_file = convert_image_to_hei(source_file_path, ".HIF")
                if source_tif:
                    log = "TIF -> HIF " + progress + ": " + output_file
                if source_jpg:
                    log = "JPG -> HIF " + progress + ": " + output_file
            if target_avif:
                output_file = convert_image_to_hei(source_file_path, ".AVIF")
                if source_tif:
                    log = "TIF -> AVIF " + progress + ": " + output_file
                if source_jpg:
                    log = "JPG -> AVIF " + progress + ": " + output_file
                
            print(log)
    
    except Exception as e:
        print("Error with exception: " + str(e))  


def convert_all_sub_folders_to_hei(source_dir, source_tif=False, source_jpg=False, 
                            target_hif=False, target_avif=False, depth=2):
    """
    Automatically run the conversion script for a given depth.
    """
    sub_dirs = tools.find_sub_dirs(source_dir, depth=depth)
    for sub_dir in sub_dirs:
        try:
            convert_image_in_dir_to_hei(sub_dir, source_tif=source_tif, source_jpg=source_jpg, 
                                 target_hif=target_hif, target_avif=target_avif)
        except Exception as e:
            print("Error:", e)




import heiya.to_hei as to_hei
import heiya.from_hei as from_hei
import heiya.extensions as extensions
import heiya.tools as tools

import os
import shutil

def preserve_original(source_dir, extension=".JPG", backup_folder_name="original"):
    """
    Take all the items with specified extension and back them up in ./original folder.
    Args:
        source_dir (str): A directory that contain files.
        extension (str): Extension of files you desire to back up of.
        backup_folder_name (str): The name of the back up folder, by default "original".
    """
    if extension in extensions.EXT_JPG:
        target_format = extensions.EXT_JPG
    elif extension in extensions.EXT_TIF:
        target_format = extensions.EXT_TIF
    elif extension in extensions.EXT_HIF:
        target_format = extensions.EXT_HIF
    elif extension in extensions.EXT_AVIF:
        target_format = extensions.EXT_AVIF
        
    original_file_list = [file for file in os.listdir(source_dir) if file.endswith(target_format) and not file.startswith("._")]
    
    backup_dir_path = os.path.join(source_dir, backup_folder_name)
    if not os.path.exists(backup_dir_path):
        os.makedirs(backup_dir_path)
    
    # Keep track of image count for displaying progress
    image_count = len(original_file_list)
    image_counter = 0
    
    for original_file in original_file_list:
        original_path = os.path.join(source_dir, original_file)
        backup_path = os.path.join(source_dir, backup_folder_name, original_file)
        output_file = shutil.copy(original_path, backup_path)
        image_counter += 1
        progress = str(image_counter) + "/" + str(image_count) + "(" + str(int((image_counter / image_count)*100)) + "%)"
        log = "Backup " + progress + ": " + output_file
        
        print(log)
        

def convert_jpg_to_he_jpg(source_dir, use_hif=False, use_avif=False, preserve_original_jpg=True):
    """
    Convert JPG to JPG using high efficiency codec.
    Args:
        source_dir (str): A directory that contain jpg files.
        use_codec (int): 0 will use HIF. 1 will use AVIF.
    """
    if not use_hif and not use_avif:
        raise ValueError("No codec is selected, please use a codec to proceed.")
    elif use_hif and use_avif:
        raise ValueError("Cannot encode using both codec, please use one codec to proceed.")
    
    if preserve_original_jpg:
        preserve_original(source_dir, extension=".JPG")
        
    if use_hif:
        to_hei.convert_image_in_dir_to_hei(source_dir, source_jpg=True, target_hif=True)
        from_hei.convert_hei_in_dir_to_jpg(source_dir, source_hif=True,fix_rotation=True)
        tools.delete_image_in_dir(source_dir, hif=True)
    elif use_avif:
        to_hei.convert_image_in_dir_to_hei(source_dir, source_jpg=True, target_avif=True)
        from_hei.convert_hei_in_dir_to_jpg(source_dir, source_avif=True,fix_rotation=True)
        tools.delete_image_in_dir(source_dir, avif=True)

    


# High Efficiency Image (AVIF/HIF) to JPG Converter 

from PIL import Image
import piexif
import pyheif

import heiya.extensions as extensions

import os
from os import listdir
from os.path import dirname, basename

def convert_hei_to_image(source_hei, target_format=".JPG", fix_rotation=True):
    """
    Convert a HEI (High Efficiency Image) file (.AVIF or .HIF) into .JPG
    
    Args:
        source_tif (str): A full file path of a .HIF image.
        target_format (str): The target format you want to convert to, by default use ".JPG"
        fix_rotation (boolean): The conversion might mess up the rotation of the image, this can help fix the issue.
    Returns:
        (str) Full file path of the generated file.
    """
    
    # Separate a full file path into directory, file name, and extension
    directory = dirname(source_hei)
    file_name  = basename(source_hei).split(".")[0] 
    extension = basename(source_hei).split(".")[1]
    
    # Read HEI file
    # Note: pillow_heif can handle both heif and avif
    hei_file = pyheif.read(source_hei)  
    
    # Construct Image object
    # Creation of image 
    image = Image.frombytes(
        hei_file.mode,
        hei_file.size,
        hei_file.data,
        "raw",
        hei_file.mode,
        hei_file.stride,
    )
    
    # Retrive the metadata
    # Reference: https://stackoverflow.com/a/65054622
    for metadata in hei_file.metadata or []:
        if metadata['type'] == 'Exif':
            exif_dict = piexif.load(metadata['data'])

    # PIL rotates the image according to exif info, so it's necessary to remove the orientation tag otherwise the image will be rotated again (1° time from PIL, 2° from viewer).
    if fix_rotation:
        exif_dict['0th'][274] = 1
        
    meta = piexif.dump(exif_dict)
    
    # Construct output file name
    output_file = directory + "/" + file_name + target_format

    # Save new image
    if meta:
        image.save(output_file, exif=meta)
    else:
        image.save(output_file)
    
    return output_file
    
    
def convert_hei_in_dir_to_image(source_dir, source_format=".AVIF", target_format=".JPG", fix_rotation=True):
    """
    Convert all the files with an extension of ".tif" into target format.
    Args:
        source_dir (str): A directory that contain tif files.
        source_format (str): The type of files to convert from. 
                            This will convert all related files (i.e. ".HIF" will convert all .HIF .HEIC .HEIF)
                            Not very elegant, to be improved in the future.
        target_format (str or tuple): The target format you want to convert to, by default use ".AVIF"
        fix_rotation (boolean): The conversion might mess up the rotation of the image, this can help fix the issue.
    """
    try:
        # Get all HIF files
        if source_format in extensions.EXT_AVIF:
            hei_format = extensions.EXT_AVIF
        elif source_format in extensions.EXT_HIF:
            hei_format = extensions.EXT_HIF
        else:
            raise ValueError("Not a valid source format.")
            
        hei_file_list = [file for file in listdir(source_dir) if file.endswith(hei_format) and not file.startswith("._")]  

        # Keep track of image count for displaying progress
        image_count = len(hei_file_list)
        image_counter = 0

        # Manipulate each tif image
        for hei_file in hei_file_list:
            hei_file_path = os.path.join(source_dir, hei_file)
            output_file = convert_hei_to_image(hei_file_path, target_format, fix_rotation=fix_rotation)

            image_counter += 1
            progress = str(image_counter) + "/" + str(image_count) + "(" + str(int((image_counter / image_count)*100)) + "%)"
            log = source_format + " -> " + target_format + " " + progress + ": " + output_file
            print(log)
    
    except Exception as e:
        print("Error with exception: " + str(e)) 
        
        
def convert_hei_in_dir_to_jpg(source_dir, source_hif=False, source_avif=False, fix_rotation=True):
    """
    Call method from the user to specify which file format in a directory to convert to JPG.
    Args:
        source_dir (str): A directory that contain tif files.
        source_hif (boolean): If true, convert all the HIF formats inside the directory into JPG
        source_avif (boolean): If true, convert all the AVIF formats inside the directory into JPG
        fix_rotation (boolean): The conversion might mess up the rotation of the image, this can help fix the issue.
    """
    if source_hif:
        convert_hei_in_dir_to_image(source_dir, source_format=".HIF", target_format=".JPG", fix_rotation=fix_rotation)
    if source_avif:
        convert_hei_in_dir_to_image(source_dir, source_format=".AVIF", target_format=".JPG", fix_rotation=fix_rotation)



from setuptools import setup, find_packages

with open('README.md') as readme_file:
    README = readme_file.read()

with open('HISTORY.md') as history_file:
    HISTORY = history_file.read()

setup_args = dict(
    name='heiya',
    version='0.0.9',
    description='Convert image to High Efficiency Image (HEIC/AVIF) and back with ease.',
    long_description_content_type="text/markdown",
    long_description=README + '\n\n' + HISTORY,
    license='Apache License 2.0',
    packages=find_packages(),
    author='Hongjun Wu',
    author_email='hw434@cornell.edu',
    keywords=['HEIC', 'AVIF', 'TIFF', "JPG",'ImageConversion'],
    url='https://github.com/wu-hongjun/heiya',
    download_url='https://pypi.org/project/heiya/'
)

install_requires = [
    'piexif',
    'pyheif',
    'pillow_heif',
    'pyperclip',
    'pathlib',
    'Pillow',
]

if __name__ == '__main__':
    setup(**setup_args, install_requires=install_requires)
